# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Bring calendars using X-WR-TIMEZONE into RFC 5545 form."""
import sys
import pytz
from icalendar.prop import vDDDTypes, vDDDLists
import datetime
import icalendar

X_WR_TIMEZONE = "X-WR-TIMEZONE"


def list_is(l1, l2):
    """Return wether all contents of two lists are identical."""
    return len(l1) == len(l2) and all(e1 is e2 for e1, e2 in zip(l1, l2))


class CalendarWalker:
    """I walk along the components and values of an icalendar object.

    The idea is the same as a visitor pattern.
    """

    VALUE_ATTRIBUTES = ['DTSTART', 'DTEND', 'RDATE', 'RECURRENCE-ID', 'EXDATE']

    def copy_if_changed(self, component, attributes, subcomponents):
        """Check if an icalendar Component has changed and copy it if it has.

        atributes and subcomponents are put into the copy."""
        for key, value in attributes.items():
            if component[key] is not value:
                return self.copy_component(component, attributes, subcomponents)
        assert len(component.subcomponents) == len(subcomponents)
        for new_subcomponent, old_subcomponent in zip(subcomponents, component.subcomponents):
            if new_subcomponent is not old_subcomponent:
                return self.copy_component(component, attributes, subcomponents)
        return component

    def copy_component(self, component, attributes, subcomponents):
        """Create a copy of the component with attributes and subcomponents."""
        component = component.copy()
        for key, value in attributes.items():
            component[key] = value
        assert len(component.subcomponents) == 0
        for subcomponent in subcomponents:
             component.add_component(subcomponent)
        return component

    def walk(self, calendar):
        """Walk along the calendar and return the changed or identical object."""
        subcomponents = []
        for subcomponent in calendar.subcomponents:
            if isinstance(subcomponent, icalendar.cal.Event):
                subcomponent = self.walk_event(subcomponent)
            subcomponents.append(subcomponent)
        return self.copy_if_changed(calendar, {}, subcomponents)

    def walk_event(self, event):
        """Walk along the event and return the changed or identical object."""
        attributes = {}
        for name in self.VALUE_ATTRIBUTES:
            value = event.get(name)
            if value is not None:
                attributes[name] = self.walk_value(value)
        return self.copy_if_changed(event, attributes, event.subcomponents)

    def walk_value_default(self, value):
        """Default method for walking along a value type."""
        return value

    def walk_value(self, value):
        """Walk along a value type."""
        name = "walk_value_" + type(value).__name__
        walk = getattr(self, name, self.walk_value_default)
        return walk(value)

    def walk_value_list(self, l):
        """Walk through a list of values."""
        v = list(map(self.walk_value, l))
        if list_is(v, l):
            return l
        return v

    def walk_value_vDDDLists(self, l):
        dts = [ddd.dt for ddd in l.dts]
        new_dts = [self.walk_value(dt) for dt in dts]
        if list_is(new_dts, dts):
            return l
        return vDDDLists(new_dts)

    def walk_value_vDDDTypes(self, value):
        """Walk along an icalendar value type"""
        dt = self.walk_value(value.dt)
        if dt is value.dt:
            return value
        return vDDDTypes(dt)

    def walk_value_datetime(self, dt):
        """Walk along a datetime.datetime object."""
        return dt

    def is_UTC(self, dt):
        """Return whether the time zone is a UTC time zone."""
        if dt.tzname() is None:
            return False
        return dt.tzname().upper() == "UTC"

    def is_Floating(self, dt):
        return dt.tzname() is None


class UTCChangingWalker(CalendarWalker):
    """Changes the UTC time zone into a new time zone."""

    def __init__(self, timezone):
        """Initialize the walker with the new time zone."""
        self.new_timezone = timezone

    def walk_value_datetime(self, dt):
        """Walk along a datetime.datetime object."""
        if self.is_UTC(dt):
            return dt.astimezone(self.new_timezone)
        elif self.is_Floating(dt):
            return dt.replace(tzinfo=self.new_timezone)
        return dt


def to_standard(calendar, timezone=None):
    """Make a calendar that might use X-WR-TIMEZONE compatible with RFC 5545.

    Arguments:
    - calendar is an icalendar.Calendar object. It does not need to have
        the X-WR-TIMEZONE property but if it has, calendar  will be converted
        to conform to RFC 5545.
    - timezone is an optional timezone argument if you want to override the
        existence of the actual X-WR-TIMEZONE property of the calendar.
        This can be a string like "Europe/Berlin" or "UTC" or a
        pytz.timezone or any other timezone accepted by the datetime module.
    """
    if timezone is None:
        timezone = calendar.get(X_WR_TIMEZONE, None)
    if timezone is not None and not isinstance(timezone, datetime.tzinfo):
        timezone = pytz.timezone(timezone)
    if timezone is not None:
        walker = UTCChangingWalker(timezone)
        return walker.walk(calendar)
    return calendar

def main():
    """x-wr-timezone converts ICSfiles with X-WR-TIMEZONE to use RFC 5545 instead.

    Convert input:

        cat in.ics | x-wr-timezone > out.ics
        wget -O- https://example.org/in.ics | x-wr-timezone > out.ics
        curl https://example.org/in.ics | x-wr-timezone > out.ics

    Convert files:

        x-wr-timezone in.ics out.ics

    Get help:

        x-wr-timezone --help

    For bug reports, code and questions, visit the projet page:

        https://github.com/niccokunzmann/x-wr-timezone

    License: LPGLv3+
    """
    if len(sys.argv) == 1:
        in_file = getattr(sys.stdin, "buffer", sys.stdin)
        out_file = getattr(sys.stdout, "buffer", sys.stdout)
    elif len(sys.argv) == 3:
        in_file = open(sys.argv[1], 'rb')
        out_file = open(sys.argv[2], 'wb')
    else:
        sys.stdout.write(main.__doc__)
        return 0
    input = in_file.read()
    calendar = icalendar.Calendar.from_ical(input)
    output = to_standard(calendar).to_ical()
    out_file.write(output)
    return 0


__all__ = [
    "main", "to_standard", "UTCChangingWalker", "list_is",
    "X_WR_TIMEZONE", "CalendarWalker"
]



"""This test converts example calendars"""
import pytest
import x_wr_timezone
import pytz


def assert_has_line(bytes, content, message):
    lines = message_lines = [line for line in bytes.decode("UTF-8").splitlines() if content[0] in line]
    for c in content:
        lines = [line for line in lines if c in line]
    assert lines, message + " One of these lines should contain {}: {}".format(content, message_lines)


def test_input_to_output(to_standard, calendar_pair):
    """Test the calendars which are a pair of input and output."""
    output = to_standard(calendar_pair.input.as_icalendar())
    assert output == calendar_pair.output.as_icalendar(), calendar_pair.message


def test_output_stays_the_same(to_standard, output_calendar):
    assert output_calendar.is_corrected_output()
    output = to_standard(output_calendar.as_icalendar())
    assert output == output_calendar.as_icalendar(), "A calendar that was modified one should stay as it is."


def RDATE(dt):
    return ("rdate-hackerpublicradio.in.ics", ("RDATE", "TZID=Europe/London", dt), "RDATE is converted with value " + dt)
def EXDATE(dt):
    return ("exdate-hackerpublicradio-modified.in.ics", ("EXDATE", "TZID=Europe/London", dt), "EXDATE is converted with value " + dt)


@pytest.mark.parametrize("calendar_name,content,message", [
# DTSTART;TZID=America/New_York:20211222T120000
# DTEND;TZID=America/New_York:20211222T130000
    ("single-events-DTSTART-DTEND.in.ics", ("DTSTART", "TZID=America/New_York", ":20211222T120000"), "DTSTART should be converted."),
    ("single-events-DTSTART-DTEND.in.ics", ("DTEND", "TZID=America/New_York", ":20211222T130000"), "DTEND should be converted."),
    ("single-events-DTSTART-DTEND.out.ics", ("DTSTART", "TZID=America/New_York", ":20211222T120000"), "DTSTART stays the same in already converted calendar."),
# DTSTART;TZID=America/New_York:20211222T210000
# DTEND;TZID=America/New_York:20211222T220000
    ("single-events-DTSTART-DTEND.in.ics", ("DTSTART", "TZID=America/New_York", ":20211222T210000"), "DTSTART should be converted."),
    ("single-events-DTSTART-DTEND.in.ics", ("DTEND", "TZID=America/New_York", ":20211222T220000"), "DTEND should be converted."),
    RDATE("20130803T200000"), # summer
    RDATE("20130831T200000"), # summer
    RDATE("20131005T200000"), # summer
    RDATE("20131102T190000"),
    RDATE("20131130T190000"),
    RDATE("20140104T190000"),
    RDATE("20140201T190000"),
    RDATE("20140301T190000"),
    RDATE("20140405T200000"), # summer
    RDATE("20140503T200000"), # summer
    RDATE("20140531T200000"), # summer
    RDATE("20140705T200000"), # summer
    # RECURRENCE-ID as well as DTSTART and DTEND
    ("moved-event-RECURRENCE-ID.in.ics", ("RECURRENCE-ID", "TZID=Europe/Berlin", "20211231T213000"), "The RECURRENCE-ID  depends on DTSTART and should therefore be converted."),
    ("moved-event-RECURRENCE-ID.in.ics", ("DTSTART", "TZID=Europe/Berlin", "20211126T21"), "DTSTART is converted."),
    ("moved-event-RECURRENCE-ID.in.ics", ("DTEND", "TZID=Europe/Berlin", "20211126T213000"), "DTEND is converted."),
    # test EXDATE
    EXDATE("20130803T200000"), # summer
    EXDATE("20130831T200000"), # summer
    EXDATE("20131005T200000"), # summer
    EXDATE("20131102T190000"),
    EXDATE("20131130T190000"),
    EXDATE("20140104T190000"),
    EXDATE("20140201T190000"),
    EXDATE("20140301T190000"),
    EXDATE("20140405T200000"), # summer
    EXDATE("20140503T200000"), # summer
    EXDATE("20140531T200000"), # summer
    EXDATE("20140705T200000"), # summer
])
def test_conversion_changes_the_time_zone(to_standard, calendars, calendar_name, content, message):
    calendar = calendars[calendar_name]
    new_calendar = to_standard(calendar.as_icalendar())
    output_bytes = new_calendar.to_ical()
    assert_has_line(output_bytes, content, message)


@pytest.mark.parametrize("calendar_name,content,message", [
    ("single-event-no-tz.in.ics", ("DTSTART", "TZID=Europe/Brussels", ":20210916T210000"), "DTSTART should be updated."),
    ("single-event-no-tz.in.ics", ("DTEND", "TZID=Europe/Brussels", ":20210916T224500"), "DTEND should be updated."),
    ("single-event-no-tz.out.ics", ("DTSTART", "TZID=Europe/Brussels", ":20210916T210000"), "DTSTART stays the same in already converted calendar."),
])
def test_conversion_adds_the_time_zone(to_standard, calendars, calendar_name, content, message):
    calendar = calendars[calendar_name]
    new_calendar = to_standard(calendar.as_icalendar())
    output_bytes = new_calendar.to_ical()
    print(output_bytes)
    assert_has_line(output_bytes, content, message)


@pytest.mark.parametrize("tz,line,message,calendar_name", [
    ("Europe/Paris", ("DTSTART", "TZID=Europe/Paris" ,"20211223T030000"), "(1) Use string as timezone", "single-events-DTSTART-DTEND.in.ics"),
    (pytz.timezone("Europe/Berlin"), ("DTSTART", "TZID=Europe/Berlin", "20211223T030000"), "(2) Use pytz.timezone as timezone", "single-events-DTSTART-DTEND.in.ics"),
    (pytz.UTC, ("DTSTART", "20211222T170000Z"), "(3) Use pytz.UTC as timezone", "single-events-DTSTART-DTEND.in.ics"),
])
def test_timezone_parameter(calendars, tz, line, message, calendar_name):
    calendar = calendars[calendar_name]
    new_calendar = x_wr_timezone.to_standard(calendar.as_icalendar(), timezone=tz)
    output_bytes = new_calendar.to_ical()
    assert_has_line(output_bytes, line, message)



"""Test the command line interfae explicitely"""
import subprocess


CMD = "x-wr-timezone"


def test_help():
    """Test that a help is being displayed."""
    help = subprocess.check_output([CMD, "--help"])
    assert b'x-wr-timezone' in help



"""Make sure the examples are what we want them to be"""

def test_calendars_are_the_same(calendars):
    assert calendars["single-event-x-wr-timezone-not-used.in"].to_ical() == \
        calendars["single-event-x-wr-timezone-not-used.out"].to_ical()


def test_calendars_differ(calendars):
    assert calendars["single-events-DTSTART-DTEND.in"].to_ical() != \
        calendars["single-events-DTSTART-DTEND.out"].to_ical()



'''
Test the README file.

This is necessary because a deployment does not work if the README file
has errors.

Credits: https://stackoverflow.com/a/47494076/1320237
'''
import os
import sys
import restructuredtext_lint

HERE = os.path.dirname(__file__)
readme_path = os.path.join(os.path.dirname(HERE), "README.rst")

def test_readme_file():
    '''CHeck README file for errors.'''
    messages = restructuredtext_lint.lint_file(readme_path)
    error_message = "expected to have no messages about the README file!"
    for message in messages:
        print(message.astext())
        error_message += "\n" + message.astext()
    assert len(messages) == 0, error_message


"""These tests check that the argument is not modified."""
from x_wr_timezone import to_standard
import pytest


@pytest.mark.parametrize("calendar_name,message", [
   ("x-wr-timezone-not-present.in.ics", "No X-WR-TIMEZONE no change"),
   ("rdate-hackerpublicradio.out.ics", "no change, same calendar"),
])
def test_no_change(calendars, calendar_name, message):
    """Test when calendars do not change."""
    calendar = calendars[calendar_name].as_icalendar()
    same_calendar = to_standard(calendar)
    assert same_calendar is calendar, message

def test_calendar_is_changed(calendars):
    """If X-WR-TIMEZONE changes the calendar, it should create a copy."""
    calendar = calendars["rdate-hackerpublicradio.in.ics"].as_icalendar()
    changed_calendar = to_standard(calendar)
    assert changed_calendar is not calendar


def test_components_are_not_altered(calendars):
    """The components of the calendar should be altered but originals left intact."""
    calendar = calendars["rdate-hackerpublicradio.in.ics"].as_icalendar()
    changed_calendar = to_standard(calendar)
    output_original = calendar.to_ical().decode("UTF-8")
    output_original_lines = output_original.splitlines()
    output_changed = changed_calendar.to_ical().decode("UTF-8")
    output_changed_lines = output_changed.splitlines()
    for original_line, new_line in zip(output_original_lines, output_changed_lines):
        print("o:\t", original_line, "\n\t", new_line)
    assert output_original_lines != output_changed_lines


def get_lines(calendar):
    output_changed = calendar.to_ical().decode("UTF-8")
    return output_changed.splitlines()


def filter_lines(calendar, content):
    return [line for line in get_lines(calendar) if content in line]


@pytest.mark.parametrize("calendar_name", [
    "rdate-hackerpublicradio.in.ics",
    "moved-event-RECURRENCE-ID.in.ics",
])
@pytest.mark.parametrize("property", "RDATE,BEGIN,BEGIN:CALENDAR,BEGIN:EVENT,END:EVENT,END:CALENDAR,DTSTART,DTEND,SUMMARY".split(","))
def test_components_all_in_there(calendars, calendar_name, property):
    """Make sure all components are there."""
    calendar = calendars[calendar_name].as_icalendar()
    l1 = filter_lines(calendar, property)
    changed_calendar = to_standard(calendar)
    l2 = filter_lines(calendar, property)
    l3 = filter_lines(changed_calendar, property)
    assert l1 == l2, "calendar itself should not have changed"
    assert len(l2) == len(l3), "no components should be added"





"""Test and fixture initialization."""
import x_wr_timezone
import icalendar
import pytest
import sys
import os
import tempfile
import shutil
import subprocess

HERE = os.path.dirname(__file__)
REPO = os.path.dirname(HERE)

sys.path.append(REPO)

CALENDARS_FOLDER = os.path.join(HERE, "calendars")

NAME_END_INPUT = ".in"
NAME_END_OUTPUT = ".out"
EXECUTABLE = 'x-wr-timezone'

example_calendars = {} # name: calendar

class TestCalendar:
    """A calendar in the test folder."""

    def __init__(self, path):
        self.path = path

    @property
    def filename(self):
        return os.path.basename(self.path)

    @property
    def filename_without_ending(self):
        return os.path.splitext(self.filename)[0]

    @property
    def name(self):
        """Return the nicely readable id of the calendar."""
        return self.filename_without_ending.replace("-", " ")

    def is_input(self):
        """Whether this is a calendar which is used as an input for to_standard()."""
        return self.name.endswith(NAME_END_INPUT)

    def is_corrected_output(self):
        """Whether this is a calendar which would be a result of to_standard()."""
        return self.name.endswith(NAME_END_OUTPUT)

    def get_corrected_output_name(self):
        assert self.is_input(), "Only input calendars can have an output."
        return self.name[:len(self.name) - len(NAME_END_INPUT)] + NAME_END_OUTPUT

    def has_corrected_output(self):
        """Whether this has a calendar resulting from to_standard()."""
        return self.get_corrected_output_name() in example_calendars

    def get_corrected_output(self):
        """Return the corrected output of to_standard() for this calendar.

        KeyError if absent."""
        return example_calendars[self.get_corrected_output_name()]

    def __repr__(self):
        return "<{}>".format(self.name)

    def as_bytes(self):
        with open(self.path, "rb") as file:
            return file.read()

    def as_icalendar(self):
        return icalendar.Calendar.from_ical(self.as_bytes())

    def to_ical(self):
        return self.as_icalendar().to_ical()

for calendar_file in os.listdir(CALENDARS_FOLDER):
    calendar_path = os.path.join(CALENDARS_FOLDER, calendar_file)
    calendar = TestCalendar(calendar_path)
    example_calendars[calendar.name] = calendar
    example_calendars[calendar.filename] = calendar
    example_calendars[calendar.filename_without_ending] = calendar

class CalendarPair:
    """A pair of input and output calendars."""

    def __init__(self, input, output):
        self.input = input
        self.output = output

    @property
    def message(self):
        return self.input.as_icalendar().get("X-WR-CALNAME",
            "Pair of {} + {}".format(self.input.name, NAME_END_OUTPUT))

    def __repr__(self):
        return "<{}+{}>".format(self.input.name, NAME_END_OUTPUT)

# for parametrizing fixtures, see https://docs.pytest.org/en/latest/fixture.html#parametrizing-fixtures
@pytest.fixture(params=[calendar for calendar in example_calendars.values() if calendar.is_input() and calendar.has_corrected_output()])
def calendar_pair(request):
    """A pair of input and output calendar examples from the test/calendars folder."""
    return CalendarPair(request.param, request.param.get_corrected_output())


@pytest.fixture(params=[calendar for calendar in example_calendars.values() if calendar.is_corrected_output()])
def output_calendar(request):
    """A TestCalendar ending with .out in from the test/calendars folder."""
    return request.param


def to_standard_cmd_stdio(calendar):
    """Use the command line and piping."""
    input = calendar.to_ical()
    process = subprocess.Popen([EXECUTABLE], stdout=subprocess.PIPE, stdin=subprocess.PIPE)
    output = process.communicate(input)[0]
    assert process.returncode == 0, "The process should not error."
    return icalendar.Calendar.from_ical(output)


def to_standard_cmd_file(calendar):
    d = tempfile.mkdtemp(prefix="pytest-")
    try:
        input = calendar.to_ical()
        in_path = os.path.join(d, "in.ics")
        out_path = os.path.join(d, "out.ics")
        with open(in_path, 'wb') as f:
            f.write(input)
        subprocess.check_call([EXECUTABLE, in_path, out_path])
        with open(out_path, 'rb') as f:
            output = f.read()
    finally:
        shutil.rmtree(d)
    return icalendar.Calendar.from_ical(output)

conversions = {
    "all": [x_wr_timezone.to_standard, to_standard_cmd_stdio, to_standard_cmd_file],
    "fast": [x_wr_timezone.to_standard],
    "io": [to_standard_cmd_stdio],
    "file": [to_standard_cmd_file],
}

@pytest.fixture(params=[
    x_wr_timezone.to_standard,
    to_standard_cmd_stdio,
    to_standard_cmd_file,
])
def to_standard(request, pytestconfig):
    """Change the to_standard() function to test several different methods.

    Use:
    - fast - use x_wr_timezone.to_standard(...)
    - io - use cat ... > x-wr-timezone
    - file - use x-wr-timezone in.ics out.ics
    - all - all of the above
    """
    to_standard = request.param
    if to_standard not in conversions[pytestconfig.option.to_standard]:
        pytest.skip("Use --x-wr-timezone=all to ativate all tests.")
    return to_standard


def pytest_addoption(parser):
    group = parser.getgroup("x-wr-timezone")
    group.addoption(
        "--x-wr-timezone",
        action="store",
        dest="to_standard",
        choices=("all", "file", "io", "fast"),
        default="fast",
        metavar="MODE",
        help=to_standard.__doc__,
    )


@pytest.fixture()
def calendars():
    """A mapping of all TestCalendars in the test/calendars folder."""
    return example_calendars.copy()


@pytest.fixture()
def todo():
    """Skip a test because it needs to be written first."""
    pytest.skip("This test is not yet implemented.")
